/*
 * Copyright 2019 Qubole, Inc.  All rights reserved.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.qubole.spark.hiveacid.util

import org.apache.hadoop.hive.common.`type`.{Timestamp => HiveTimestamp}
import org.apache.spark.internal.Logging
import org.apache.spark.sql.catalyst.util.DateTimeUtils
import org.apache.spark.sql.catalyst.util.DateTimeUtils.{MICROS_PER_MILLIS, NANOS_PER_MICROS}

import java.util.TimeZone

object HWCDateTimeUtils extends Logging {

  /**
   * Adjusts input millis to destination timezone.
   * <p>
   * Similar to the code in ORC project -
   * https://github.com/apache/orc/blob/rel/release-1.6.7/java/core/src/java/org/apache/orc/impl/SerializationUtils.java#L1395-L1417
   *
   * @param sourceTz      the timezone we are moving from
   * @param destinationTz the timezone we are moving to
   * @param millis        the point in time
   * @return modified milliseconds according to destination timezone
   */
  def convertMillis(millis: Long, sourceTz: TimeZone, destinationTz: TimeZone): Long = {
    val sourceOffset = sourceTz.getOffset(millis)
    val destinationOffset = destinationTz.getOffset(millis)
    val adjustedMillis = millis + sourceOffset - destinationOffset
    // If the timezone adjustment moves the millis across a DST boundary, we
    // need to reevaluate the offsets.
    val dstAdjustedOffset = sourceOffset - destinationTz.getOffset(adjustedMillis)
    // final millis after adjusting the offsets
    millis + dstAdjustedOffset
  }

  /**
   * Adjusts input micros to destination timezone.
   * <p>
   * Similar to the code in ORC project -
   * https://github.com/apache/orc/blob/rel/release-1.6.7/java/core/src/java/org/apache/orc/impl/SerializationUtils.java#L1395-L1417
   *
   * @param sourceTz      the timezone we are moving from
   * @param destinationTz the timezone we are moving to
   * @param micros        the point in time
   * @return modified micros according to destination timezone
   */
  def convertMicros(micros: Long, sourceTz: TimeZone, destinationTz: TimeZone): Long = {
    val millis = micros / 1000
    val remainingMicros = micros % 1000
    convertMillis(millis, sourceTz, destinationTz) * 1000 + remainingMicros
  }

  /**
   *
   * Converts hive Timestamp to JVM timezone,
   * this method assumes that the input hiveTimestamp is always in UTC
   *
   * @param hiveTimestamp
   * @return converted micros
   */
  def hiveTimestampToMicrosForDisplay(hiveTimestamp: HiveTimestamp): Long = {
    val epochMicros = hiveTimestamp.toEpochMilli * MICROS_PER_MILLIS
    +(hiveTimestamp.getNanos / NANOS_PER_MICROS) % NANOS_PER_MICROS
    convertMicros(epochMicros, DateTimeUtils.TimeZoneUTC, DateTimeUtils.defaultTimeZone())
  }
  /**
   *
   * Converts micros from JVM timezone to UTC and returns hive timestamp
   * this method assumes that the input micros is always in JVM timezone
   *
   * @param hiveTimestamp
   * @return converted micros
   */
  def microsToHiveTimestampForWriting(micros: Long): HiveTimestamp = {
    val utcMicros = convertMicros(micros, DateTimeUtils.defaultTimeZone(), DateTimeUtils.TimeZoneUTC)
    val millis: Long = utcMicros / 1000
    val nanos: Long = (utcMicros % 1000) * 1000
    HiveTimestamp.ofEpochMilli(millis, nanos.toInt)
  }

}
